from __future__ import absolute_import, division, print_function
import numpy as np
from collections import OrderedDict
from sklearn.metrics import auc, log_loss, precision_recall_curve, roc_auc_score, average_precision_score


def loss(labels, predictions):
    return log_loss(labels, predictions)


def positive_accuracy(labels, predictions, threshold=0.5):
    return 100 * (predictions[labels] > threshold).mean()


def negative_accuracy(labels, predictions, threshold=0.5):
    return 100 * (predictions[~labels] < threshold).mean()


def balanced_accuracy(labels, predictions, threshold=0.5):
    return (positive_accuracy(labels, predictions, threshold) +
            negative_accuracy(labels, predictions, threshold)) / 2


def auROC(labels, predictions):
    return roc_auc_score(labels, predictions)

#handles ties properly
def auPRC_careful(labels, predictions):
    auc_careful = average_precision_score(labels, predictions)
    return auc_careful

#may not handle ties properly 
def auPRC_trapezoid(labels, predictions):
    precision_trapezoid,recall_trapezoid=precision_recall_curve(labels,predictions)[:2]
    auc_trapezoid=auc(recall_trapezoid,precision_trapezoid) 
    return auc_trapezoid



def recall_at_precision_threshold(labels, predictions,precision_threshold):
    precision, recall = precision_recall_curve(labels, predictions)[:2]
    return 100 * recall[np.searchsorted(precision - precision_threshold, 0)]


class ClassificationResult(object):

    def __init__(self, labels, predictions, task_names=None):
        assert labels.dtype == bool
        self.results = [OrderedDict((
            ('Loss', loss(task_labels, task_predictions)),
            ('Balanced accuracy', balanced_accuracy(
                task_labels, task_predictions)),
            ('auROC', auROC(task_labels, task_predictions)),
            ('auPRC Careful', auPRC_careful(task_labels, task_predictions)),
            ('auPRC Trapezoid', auPRC_trapezoid(task_labels,task_predictions)),
            ('Recall at 5% FDR', recall_at_precision_threshold(
                task_labels, task_predictions, 0.95)),
            ('Recall at 10% FDR', recall_at_precision_threshold(
                task_labels, task_predictions, 0.9)),
            ('Recall at 20% FDR', recall_at_precision_threshold(
                task_labels, task_predictions, 0.8)),
            ('Num Positives', task_labels.sum()),
            ('Num Negatives', (1 - task_labels).sum())
        )) for task_labels, task_predictions in zip(labels.T, predictions.T)]
        self.task_names = task_names
        self.multitask = labels.shape[1] > 1

    def __str__(self):
        return '\n'.join(
            '{}Loss: {:.4f}\tBalanced Accuracy: {:.2f}%\t '
            'auROC: {:.3f}\t auPRC Careful: {:.3f}\t auPRC Trapezoidal: {:.3f}\n\t'
            'Recall at 5%|10%|20% FDR: {:.1f}%|{:.1f}%|{:.1f}%\t '
            'Num Positives: {}\t Num Negatives: {}'.format(
                '{}: '.format('Task {}'.format(
                    self.task_names[task_index]
                    if self.task_names is not None else task_index))
                if self.multitask else '', *results.values())
            for task_index, results in enumerate(self.results))

    def __getitem__(self, item):
        return np.array([task_results[item] for task_results in self.results])



import argparse
import numpy as np, random
np.random.seed(1)
random.seed(1)
from .utils import encode_fasta_sequences, get_sequence_strings
try:
    from sklearn.model_selection import train_test_split  # sklearn >= 0.18
except ImportError:
    from sklearn.cross_validation import train_test_split  # sklearn < 0.18

def parse_args():
    parser = argparse.ArgumentParser(
        description='main script for DragoNN modeling of sequence data.')
    # define parsers with common arguments
    fasta_pair_parser = argparse.ArgumentParser(add_help=False)
    fasta_pair_parser.add_argument('--pos-sequences', type=str, required=True,
                                   help='fasta with positive sequences')
    fasta_pair_parser.add_argument('--neg-sequences', type=str, required=True,
                                   help='fasta with negative sequences')
    fasta_pair_parser.add_argument('--pos-validation-sequences',type=str,required=False,
                                   help='validation set fasta with positive sequences')
    fasta_pair_parser.add_argument('--neg-validation-sequences',type=str,required=False)
    single_fasta_parser = argparse.ArgumentParser(add_help=False)
    single_fasta_parser.add_argument('--sequences', type=str, required=True,
                                    help='fasta with sequences')
    prefix_parser = argparse.ArgumentParser(add_help=False)
    prefix_parser.add_argument('--prefix', type=str, required=True,
                               help='prefix to output files')
    model_files_parser = argparse.ArgumentParser(add_help=False)
    model_files_parser.add_argument('--arch-file', type=str, required=True,
                                    help='model json file')
    model_files_parser.add_argument('--weights-file', type=str, required=True,
                                    help='weights hd5 file')
    sequence_dnn_args_parser = argparse.ArgumentParser(add_help=False)
    sequence_dnn_args_parser.add_argument('--num-filters', type=int, nargs="+")
    sequence_dnn_args_parser.add_argument('--conv-width', type=int, nargs="+")
    sequence_dnn_args_parser.add_argument('--pool-width', type=int)
    sequence_dnn_args_parser.add_argument('--L1', type=float)
    sequence_dnn_args_parser.add_argument('--dropout', type=float)
    sequence_dnn_args_parser.add_argument('--verbose', type=int)
    # define commands 
    subparsers = parser.add_subparsers(help='dragonn command help', dest='command')
    subparsers.required = True #http://bugs.python.org/issue9253#msg186387
    train_parser = subparsers.add_parser('train',
                                         parents=[fasta_pair_parser,
                                                  prefix_parser,
                                                  sequence_dnn_args_parser],
                                         help='model training help')
    train_parser.add_argument('--arch-file', type=str, required=False, default=None,
                              help='model json file')
    train_parser.add_argument('--weights-file', type=str, required=False, default=None,
                              help='weights hd5 file')
    train_parser.add_argument('--model-hdf5-file',type=str,required=False, default=None,
                              help='hdf5 file containing keras model')

    test_parser = subparsers.add_parser('test',
                                        parents=[fasta_pair_parser, model_files_parser],
                                        help='model testing help')
    predict_parser = subparsers.add_parser('predict',
                                           parents=[single_fasta_parser, model_files_parser],
                                           help='model prediction help')
    predict_parser.add_argument('--output-file', type=str, required=True,
                               help='output file for model predictions')
    interpret_parser = subparsers.add_parser('interpret',
                                             parents=[single_fasta_parser, model_files_parser, prefix_parser],
                                             help='model interpretation help')
    interpret_parser.add_argument('--pos-threshold', type=int, default=0.5,
                               help='Only examples with predicted positive class probability above this get interpreted. Default: 0.5')
    # return command and command arguments
    args = vars(parser.parse_args())
    command = args.pop("command", None)
    if command == "train": # check for valid model and weights files
        if args["arch_file"] is None and args["weights_file"] is not None:
            parser.error("You must provide an --arch-file corresponding to the provided --weights-file! Exiting!")
    return command, args


def main_train(pos_sequences=None,
               neg_sequences=None,
               pos_validation_sequences=None,
               neg_validation_sequences=None,
               prefix=None,
               arch_file=None,
               weights_file=None,
               **kwargs):
    kwargs = {key: value for key, value in kwargs.items() if value is not None}
    # encode fastas
    print("loading sequence data...")
    X_pos = encode_fasta_sequences(pos_sequences)
    y_pos = np.array([[True]]*len(X_pos))
    X_neg = encode_fasta_sequences(neg_sequences)
    y_neg = np.array([[False]]*len(X_neg))
    X = np.concatenate((X_pos, X_neg))
    y = np.concatenate((y_pos, y_neg))
    #if a validation set is provided by the user, encode that as well
    if (pos_validation_sequences!=None or neg_validation_sequences!=None):
        #both positive and negative validation sequences must be provided. 
        assert neg_validation_sequences!=None
        assert pos_validation_sequences!=None
        X_valid_pos=encode_fasta_sequences(pos_validation_sequences)
        X_valid_neg=encode_fasta_sequences(neg_validation_sequences)
        y_valid_pos=np.array([[True]])*len(X_valid_pos)
        y_valid_neg=np.array([[False]])*len(X_valid_neg)
        X_valid=np.concatenate((X_valid_pos, X_valid_neg))
        y_valid=np.concatenate((y_valid_pos, y_valid_neg))
    else:
        X_train, X_valid, y_train, y_valid = train_test_split(X, y, test_size=0.2)
    if arch_file is not None: # load  model
        print("loading model...")
        model = SequenceDNN.load(model_hdf5_file, arch_file, weights_file)
    else: # initialize model
        print("initializing model...")
        model = SequenceDNN(seq_length=X_train.shape[-1], **kwargs)
    # train
    print("starting model training...")
    model.train(X_train, y_train, validation_data=(X_valid, y_valid))
    valid_result = model.test(X_valid, y_valid)
    print("final validation metrics:")
    print(valid_result)
    # save
    print("saving model files..")
    model.save(prefix)
    print("Done!")


def main_test(pos_sequences=None,
              neg_sequences=None,
              arch_file=None,
              weights_file=None):
    # encode fastas
    print("loading sequence data...")
    X_test_pos = encode_fasta_sequences(pos_sequences)
    y_test_pos = np.array([[True]]*len(X_test_pos))
    X_test_neg = encode_fasta_sequences(neg_sequences)
    y_test_neg = np.array([[False]]*len(X_test_neg))
    X_test = np.concatenate((X_test_pos, X_test_neg))
    y_test = np.concatenate((y_test_pos, y_test_neg))
    # load model
    print("loading model...")
    model = SequenceDNN.load(arch_file, weights_file)
    # test
    print("testing model...")
    test_result = model.test(X_test, y_test)
    print(test_result)


def main_predict(sequences=None,
                 arch_file=None,
                 weights_file=None,
                 output_file=None):
    # encode fasta
    print("loading sequence data...")
    X = encode_fasta_sequences(sequences)
    # load model
    print("loading model...")
    model = SequenceDNN.load(arch_file, weights_file)
    # predict
    print("getting predictions...")
    predictions = model.predict(X)
    # save predictions
    print("saving predictions to output file...")
    np.savetxt(output_file, predictions)
    print("Done!")


def main_interpret(sequences=None,
                   arch_file=None,
                   weights_file=None,
                   pos_threshold=None,
                   peak_width=10,
                   prefix=None):
    # encode fasta
    print("loading sequence data...")
    X = encode_fasta_sequences(sequences)
    # load model
    print("loading model...")
    model = SequenceDNN.load(arch_file, weights_file)
    # predict
    print("getting predictions...")
    predictions = model.predict(X)
    # deeplift
    print("getting deeplift scores...")
    deeplift_scores = model.deeplift(X)
    # get important sequences and write to file
    print("extracting important sequences and writing to file...")
    for task_index, task_scores in enumerate(deeplift_scores):
        peak_positions = []
        peak_sequences = []
        for sequence_index, sequence_scores in enumerate(task_scores):
            if predictions[sequence_index, task_index] > pos_threshold:
                #print(sequence_scores.shape)
                basewise_sequence_scores = sequence_scores.max(axis=(0,1))
                peak_position = basewise_sequence_scores.argmax()
                peak_positions.append(peak_position)
                peak_sequences.append(X[sequence_index : sequence_index + 1,
                                        :,
                                        :,
                                        peak_position - peak_width :
                                        peak_position + peak_width])
            else:
                peak_positions.append(-1)
                peak_sequences.append(np.zeros((1, 1, 4, 2 * peak_width)))
        peak_sequences = np.concatenate(peak_sequences)
        peak_sequence_strings = get_sequence_strings(peak_sequences)
        # write important sequences to file
        ofname = "%s.task_%i.important_sequences.txt" % (prefix, task_index)
        with open(ofname, "w") as wf:
            for i, peak_position in enumerate(peak_positions):
                wf.write("> sequence_%i\n" % (i))
                wf.write("%i: %s\n" %(peak_position, peak_sequence_strings[i]))
    print("Done!")


def main():
    command_functions = {'train': main_train,
                         'test': main_test,
                         'predict': main_predict,
                         'interpret': main_interpret}
    command, args = parse_args()
    global SequenceDNN
    from dragonn.models import SequenceDNN
    import warnings
    warnings.filterwarnings('ignore')
    command_functions[command](**args)


if __name__ == "__main__":
    main()



from __future__ import division

import random
random.seed(1)
import numpy as np
np.random.seed(1)

from collections import namedtuple, defaultdict, OrderedDict
from simdna import simulations
from simdna.synthetic import StringEmbeddable
from dragonn.utils import get_motif_scores, one_hot_encode
from sklearn.model_selection import train_test_split 

Data = namedtuple('Data', ('X_train', 'X_valid', 'X_test',
                           'train_embeddings', 'valid_embeddings', 'test_embeddings',
                           'y_train', 'y_valid', 'y_test',
                           'motif_names'))


def get_available_simulations():
    return [function_name for function_name in dir(simulations)
            if "simulate" in function_name]


def print_available_simulations():
    for function_name in get_available_simulations():
        print(function_name)


def get_simulation_function(simulation_name):
    if simulation_name in get_available_simulations():
        return getattr(simulations, simulation_name)
    else:
        print("%s is not available. Available simulations are:" % (simulation_name))
        print_available_simulations()


def print_simulation_info(simulation_name):
    simulation_function = get_simulation_function(simulation_name)
    if simulation_function is not None:
        print(simulation_function.__doc__)


def get_simulation_data(simulation_name, simulation_parameters,
                        test_set_size=4000, validation_set_size=3200):
    simulation_function = get_simulation_function(simulation_name)
    sequences, y, embeddings = simulation_function(**simulation_parameters)
    if simulation_name == "simulate_heterodimer_grammar":
        motif_names = [simulation_parameters["motif1"],
                       simulation_parameters["motif2"]]
    elif simulation_name == "simulate_multi_motif_embedding":
        motif_names = simulation_parameters["motif_names"]
    else:
        motif_names = [simulation_parameters["motif_name"]]

    train_sequences, test_sequences, train_embeddings, test_embeddings, y_train, y_test = \
        train_test_split(sequences, embeddings, y, test_size=test_set_size)
    train_sequences, valid_sequences, train_embeddings, valid_embeddings, y_train, y_valid = \
        train_test_split(train_sequences, train_embeddings, y_train, test_size=validation_set_size)
    X_train = one_hot_encode(train_sequences)
    X_valid = one_hot_encode(valid_sequences)
    X_test = one_hot_encode(test_sequences)

    return Data(X_train, X_valid, X_test, train_embeddings, valid_embeddings, test_embeddings,
                y_train, y_valid, y_test, motif_names)





from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import warnings
import numpy as np

from keras.utils.data_utils import Sequence
from keras.utils.data_utils import OrderedEnqueuer
from keras.utils.generic_utils import Progbar
from keras.utils.generic_utils import to_list
from keras.utils.generic_utils import unpack_singleton
from keras import callbacks as cbks


def dragonn_predict_generator(model, generator,
                              steps=None,
                              max_queue_size=10,
                              workers=1,
                              use_multiprocessing=False,
                              verbose=1):
    """See docstring for `Model.predict_generator`."""
    model._make_predict_function()
    generator_indices=generator.indices
    prediction_indices=None
    batch_size=generator.batch_size 
    steps_done = 0
    wait_time = 0.01
    all_outs = []
    steps=len(generator)
    enqueuer = OrderedEnqueuer(
        generator,
        use_multiprocessing=use_multiprocessing)
    enqueuer.start(workers=workers, max_queue_size=max_queue_size)
    output_generator = enqueuer.get()
    if verbose == 1:
        progbar = Progbar(target=steps)
    try:
        while steps_done < steps:
            generator_output = next(output_generator)
            #print("got batch") 
            if isinstance(generator_output, tuple):
                # Compatibility with the generators
                # used for training.
                if len(generator_output) == 2:
                    x, idx = generator_output
                elif len(generator_output) == 3:
                    x, y, idx = generator_output
                else:
                    raise ValueError('Output of generator should be '
                                     'a tuple `(x, y, idx)` '
                                     'or `(x, idx)`. Found: ' +
                                     str(generator_output))
            else:
                raise ValueError('Output of generator should be '
                                 'a tuple `(x, y, idx)` '
                                 'or `(x, idx)`. Found: ' +
                                 str(generator_output))
            outs = model.predict_on_batch(x)
            cur_inds=generator_indices[idx*batch_size:(idx+1)*batch_size]
            if prediction_indices is None:
                prediction_indices=cur_inds
            else:
                prediction_indices=np.concatenate((prediction_indices,cur_inds),axis=0)
            outs = to_list(outs)
            if not all_outs:
                for out in outs:
                    all_outs.append([])

            for i, out in enumerate(outs):
                all_outs[i].append(out)
            steps_done += 1
            if verbose == 1:
                progbar.update(steps_done)
    except:
        print("Error, stopping enqueuer") 
        enqueuer.stop()
        print("exiting")
    enqueuer.stop()
    
    if len(all_outs) == 1:
        if steps_done == 1:
            return (all_outs[0][0],prediction_indices)
        else:
            return (np.concatenate(all_outs[0]),prediction_indices)
    if steps_done == 1:
        return ([out[0] for out in all_outs],prediction_indices)
    else:
        return ([np.concatenate(out) for out in all_outs],prediction_indices)



import keras.backend as K
import tensorflow as tf 
import pandas as pd 

#ambiguous values are indicated with =np.nan
def ambig_mean_squared_error(y_true, y_pred):
    nonAmbig=tf.math.logical_not(tf.is_nan(y_true))
    return K.mean(K.square(tf.boolean_mask(y_pred,nonAmbig) - tf.boolean_mask(y_true,nonAmbig)), axis=-1)

def ambig_binary_crossentropy(y_true,y_pred):
    nonAmbig=tf.math.logical_not(tf.is_nan(y_true))
    return K.mean(K.binary_crossentropy(tf.boolean_mask(y_true,nonAmbig), tf.boolean_mask(y_pred,nonAmbig)), axis=-1);



from abstention.calibration import PlattScaling, IsotonicRegression
import numpy as np

def get_calibration_function_classification(logits,labels):
    #need to generate separate calibration function for each task
    num_tasks=logits.shape[1]
    calibration_funcs=[]
    for task_index in range(num_tasks):
        calibration_funcs.append(PlattScaling()(
            valid_preacts=logits[:,task_index],
            valid_labels=labels.iloc[:,task_index].astype(bool)))
    return calibration_funcs

def get_calibration_function_regression(preacts,labels):
        #need to generate separate calibration function for each task
        num_tasks=preacts.shape[1]
        calibration_funcs=[]
        for task_index in range(num_tasks):
            calibration_funcs.append(IsotonicRegression()(
                valid_preacts=preacts[:,task_index],
                valid_labels=labels.iloc[:,task_index]))
        return calibration_funcs


def get_calibrated_predictions(preacts,calibration_functions):
    num_tasks=preacts.shape[1]
    num_calibration_functions=len(calibration_functions)
    assert num_tasks == num_calibration_functions
    calibrated_predictions=None
    for i in range(num_tasks):
        task_calibrated_predictions=np.expand_dims(calibration_functions[i](preacts[:,i]),axis=1)
        if calibrated_predictions is None:
            calibrated_predictions=task_calibrated_predictions
        else:
            calibrated_predictions=np.concatenate((calibrated_predictions,task_calibrated_predictions),axis=1)
    print(calibrated_predictions.shape)
    return calibrated_predictions 



from keras.callbacks import Callback
from dragonn.metrics import *
import warnings
warnings.filterwarnings('ignore')

class MetricsCallback(Callback):
        def __init__(self, train_data, validation_data):
            super().__init__()
            self.validation_data = validation_data
            self.train_data = train_data    
        def on_epoch_end(self, epoch, logs={}):            
            X_train = self.train_data[0]
            y_train = self.train_data[1]
            
            X_val = self.validation_data[0]
            y_val = self.validation_data[1]

            y_train_pred=self.model.predict(X_train)
            y_val_pred=self.model.predict(X_val)

            train_classification_result=ClassificationResult(y_train,y_train_pred)
            val_classification_result=ClassificationResult(y_val,y_val_pred)
            print("Training Data:") 
            print(train_classification_result)
            print("Validation Data:") 
            print(val_classification_result)
            
            
                                                                                    
                                                                                    



from keras.utils import Sequence
import pandas as pd
import numpy as np
import random
import math 
import pysam
from dragonn.utils import ltrdict
import threading 

def dinuc_shuffle(seq):
    #get list of dinucleotides
    nucs=[]
    for i in range(0,len(seq),2):
        nucs.append(seq[i:i+2])
    #generate a random permutation
    random.shuffle(nucs)
    return ''.join(nucs) 


def revcomp(seq):
    seq=seq[::-1].upper()
    comp_dict=dict()
    comp_dict['A']='T'
    comp_dict['T']='A'
    comp_dict['C']='G'
    comp_dict['G']='C'
    rc=[]
    for base in seq:
        if base in comp_dict:
            rc.append(comp_dict[base])
        else:
            rc.append(base)
    return ''.join(rc)


def open_data_file(data_path,tasks,num_to_read=None):
    if data_path.endswith('.hdf5'):
        if (tasks is None) and (num_to_read is not None):
            data=pd.read_hdf(data_path,start=0,stop=num_to_read)
        elif (tasks is not None) and (num_to_read is None):
            data=pd.read_hdf(data_path,columns=tasks)
        elif (tasks is None) and (num_to_read is None):
            data=pd.read_hdf(data_path)
        else: 
            data=pd.read_hdf(data_path,columns=tasks,start=0,stop=num_to_read)
    else:
        #treat as bed file 
        if (tasks is None) and (num_to_read is not None):
            data=pd.read_csv(data_path,header=0,sep='\t',index_col=[0,1,2],start=0,stop=num_to_read)
        elif (tasks is None) and (num_to_read is None):
            data=pd.read_csv(data_path,header=0,sep='\t',index_col=[0,1,2])
        else:
            data=pd.read_csv(data_path,header=0,sep='\t',nrows=1)
            chrom_col=data.columns[0]
            start_col=data.columns[1]
            end_col=data.columns[2]
            if num_to_read is None:
                data=pd.read_csv(data_path,header=0,sep='\t',usecols=[chrom_col,start_col,end_col]+tasks,index_col=[0,1,2])
            else:
                data=pd.read_csv(data_path,header=0,sep='\t',usecols=[chrom_col,start_col,end_col]+tasks,index_col=[0,1,2],start=0,stop=num_to_read)
    return data 

#use wrappers for keras Sequence generator class to allow batch shuffling upon epoch end
class DataGenerator(Sequence):
    def __init__(self,data_path,ref_fasta,batch_size=128,add_revcomp=True,tasks=None,shuffled_ref_negatives=False,upsample=True,upsample_ratio=0.1,upsample_thresh=0.5,num_to_read=None):
        self.lock = threading.Lock()        
        self.batch_size=batch_size
        #decide if reverse complement should be used
        self.add_revcomp=add_revcomp
        if add_revcomp==True:
            self.batch_size=int(batch_size/2)

        #determine whether negative set should consist of the shuffled refs.
        # If so, split batch size in 2, as each batch will be augmented with shuffled ref negatives
        # in ratio equal to positives 
        self.shuffled_ref_negatives=shuffled_ref_negatives
        if self.shuffled_ref_negatives==True:
            self.batch_size=int(self.batch_size/2)
            
        #open the reference file
        self.ref_fasta=ref_fasta
        self.data=open_data_file(data_path,tasks,num_to_read)
        
        self.indices=np.arange(self.data.shape[0])
        num_indices=self.indices.shape[0]
        self.add_revcomp=add_revcomp
        
        #set variables needed for upsampling the positives
        self.upsample=upsample
        if self.upsample==True:
            self.upsample_ratio=upsample_ratio
            self.upsample_thresh=upsample_thresh
            self.ones = self.data.loc[(self.data >= self.upsample_thresh).any(axis=1)]
            self.zeros = self.data.loc[(self.data < self.upsample_thresh).all(axis=1)]
            self.pos_batch_size = int(self.batch_size * self.upsample_ratio)
            self.neg_batch_size = self.batch_size - self.pos_batch_size
            self.pos_indices=np.arange(self.ones.shape[0])
            self.neg_indices=np.arange(self.zeros.shape[0])
            
            #wrap the positive and negative indices to reach size of self.indices
            num_pos_wraps=math.ceil(num_indices/self.pos_indices.shape[0])
            num_neg_wraps=math.ceil(num_indices/self.neg_indices.shape[0])
            self.pos_indices=np.repeat(self.pos_indices,num_pos_wraps)[0:num_indices]
            np.random.shuffle(self.pos_indices)
            self.neg_indices=np.repeat(self.neg_indices,num_neg_wraps)[0:num_indices]
            np.random.shuffle(self.neg_indices)
            
    def __len__(self):
        return math.ceil(self.data.shape[0]/self.batch_size)

    def __getitem__(self,idx):
        with self.lock:
            self.ref=pysam.FastaFile(self.ref_fasta)
            if self.shuffled_ref_negatives==True:
                return self.get_shuffled_ref_negatives_batch(idx)
            elif self.upsample==True:
                return self.get_upsampled_positives_batch(idx)
            else:
                return self.get_basic_batch(idx) 
        
    def get_shuffled_ref_negatives_batch(self,idx): 
        #get seq positions
        inds=self.indices[idx*self.batch_size:(idx+1)*self.batch_size]
        bed_entries=self.data.index[inds]
        #get sequences
        seqs=[self.ref.fetch(i[0],i[1],i[2]) for i in bed_entries]
        if self.add_revcomp==True:
            #add in the reverse-complemented sequences for training.
            seqs_rc=[revcomp(s) for s in seqs]
            seqs=seqs+seqs_rc
            
        #generate the corresponding negative set by dinucleotide-shuffling the sequences
        seqs_shuffled=[dinuc_shuffle(s) for s in seqs]
        seqs=seqs+seqs_shuffled
        #one-hot-encode the fasta sequences 
        seqs=np.array([[ltrdict.get(x,[0,0,0,0]) for x in seq] for seq in seqs])
        x_batch=np.expand_dims(seqs,1)
        y_batch=np.asarray(self.data.iloc[inds])
        if self.add_revcomp==True:
            y_batch=np.concatenate((y_batch,y_batch),axis=0)
        y_shape=y_batch.shape 
        y_batch=np.concatenate((y_batch,np.zeros(y_shape)))
        return (x_batch,y_batch)

    def get_upsampled_positives_batch(self,idx):
        #get seq positions
        pos_inds=self.pos_indices[idx*self.pos_batch_size:(idx+1)*self.pos_batch_size]
        pos_bed_entries=self.ones.index[pos_inds]
        neg_inds=self.neg_indices[idx*self.neg_batch_size:(idx+1)*self.neg_batch_size]
        neg_bed_entries=self.zeros.index[neg_inds]
    
        #print(neg_inds[0:10])
        #bed_entries=pos_bed_entries+neg_bed_entries

        #get sequences
        pos_seqs=[self.ref.fetch(i[0],i[1],i[2]) for i in pos_bed_entries]
        neg_seqs=[self.ref.fetch(i[0],i[1],i[2]) for i in neg_bed_entries]
        seqs=pos_seqs+neg_seqs 
        if self.add_revcomp==True:
            #add in the reverse-complemented sequences for training.
            seqs_rc=[revcomp(s) for s in seqs]
            seqs=seqs+seqs_rc
            
        #one-hot-encode the fasta sequences 
        seqs=np.array([[ltrdict.get(x,[0,0,0,0]) for x in seq] for seq in seqs])
        x_batch=np.expand_dims(seqs,1)
        
        #extract the positive and negative labels at the current batch of indices
        y_batch_pos=self.ones.iloc[pos_inds]
        y_batch_neg=self.zeros.iloc[neg_inds]
        y_batch=np.concatenate((y_batch_pos,y_batch_neg),axis=0)
        #add in the labels for the reverse complement sequences, if used 
        if self.add_revcomp==True:
            y_batch=np.concatenate((y_batch,y_batch),axis=0)
        return (x_batch,y_batch)            
    
    def get_basic_batch(self,idx):
        #get seq positions
        inds=self.indices[idx*self.batch_size:(idx+1)*self.batch_size]
        bed_entries=self.data.index[inds]
        #get sequences
        seqs=[self.ref.fetch(i[0],i[1],i[2]) for i in bed_entries]
        if self.add_revcomp==True:
            #add in the reverse-complemented sequences for training.
            seqs_rc=[revcomp(s) for s in seqs]
            seqs=seqs+seqs_rc
        #one-hot-encode the fasta sequences 
        seqs=np.array([[ltrdict.get(x,[0,0,0,0]) for x in seq] for seq in seqs])
        x_batch=np.expand_dims(seqs,1)
        #extract the labels at the current batch of indices 
        y_batch=np.asarray(self.data.iloc[inds])
        #add in the labels for the reverse complement sequences, if used 
        if self.add_revcomp==True:
            y_batch=np.concatenate((y_batch,y_batch),axis=0)
        return (x_batch,y_batch)    
    
    def on_epoch_end(self):
        #if upsampling is being used, shuffle the positive and negative indices 
        if self.upsample==True:
            np.random.shuffle(self.pos_indices)
            np.random.shuffle(self.neg_indices)
        else:
            np.random.shuffle(self.indices)

#generate batches of bQTL data with specified allele column name and flank size 
class BQTLGenerator(DataGenerator):
    def __init__(self,data_path,ref_fasta,allele_col,flank_size=500,batch_size=128):
        DataGenerator.__init__(self,data_path,ref_fasta,batch_size,add_revcomp=False,upsample=False)
        self.allele_col=allele_col
        self.flank_size=flank_size 

    #override the get_basic_batch definition to extract flanks around the bQTL
    # and insert the specified allele 
    def get_basic_batch(self,idx): 
        #get seq positions
        inds=self.indices[idx*self.batch_size:(idx+1)*self.batch_size]
        
        entries=self.data.iloc[inds]
        seqs=[]
        for index,row in entries.iterrows():
            allele=row[self.allele_col]
            chrom=index[0]
            pos=index[1]
            left_flank_start=(pos-1)-self.flank_size
            left_flank_end=(pos-1)
            right_flank_start=pos
            right_flank_end=(pos+self.flank_size-1)
            left_seq=self.ref.fetch(chrom,left_flank_start,left_flank_end)
            right_seq=self.ref.fetch(chrom,right_flank_start,right_flank_end)
            seq=left_seq+allele+right_seq
            seqs.append(seq)
            
        #one-hot-encode the fasta sequences 
        seqs=np.array([[ltrdict.get(x,[0,0,0,0]) for x in seq] for seq in seqs])
        x_batch=np.expand_dims(seqs,1)
        return x_batch

#Returns batch index in addition to x,y labels. Useful for prediction   
class IndexedBatchGenerator(DataGenerator):
    def __init__(self,data_path,ref_fasta,batch_size=128,add_revcomp=True,tasks=None,shuffled_ref_negatives=False,upsample=True,upsample_ratio=0.1,upsample_thresh=0.5):
        DataGenerator.__init__(self,data_path,ref_fasta,batch_size,add_revcomp=False,upsample=False)

    def __getitem__(self,idx):
        with self.lock:
            self.ref=pysam.FastaFile(self.ref_fasta)
            if self.shuffled_ref_negatives==True:
                (x_batch,y_batch)=self.get_shuffled_ref_negatives_batch(idx)
                return (x_batch,y_batch,idx)
            elif self.upsample==True:
                (x_batch,y_batch)=self.get_upsampled_positives_batch(idx)
                return (x_batch,y_batch,idx) 
            else:
                (x_batch,y_batch)=self.get_basic_batch(idx)
                return (x_batch,y_batch,idx)
            
    def on_epoch_end(self):
        return
    



from setuptools import setup,find_packages

config = {
    'include_package_data': True,
    'description': 'Deep RegulAtory GenOmic Neural Networks (DragoNN)',
    'version': '0.4.3-c',
    'packages': ['dragonn'],
    'setup_requires': [],
    'install_requires': ['numpy>=1.16', 'keras==2.3.0','tensorflow==1.15','deeplift>=0.6.9.0', 'shapely', 'matplotlib>=3.1.1', 'plotnine>=0.4.0','scikit-learn>=0.20.0', 'pydot_ng>=1.0.0', 'h5py','seqdataloader>=0.124','simdna_dragonn','abstention'],
    'dependency_links': [],
    'scripts': [],
    'entry_points': {'console_scripts': ['dragonn = dragonn.__init__:main']},
    'name': 'dragonn'
}

if __name__== '__main__':
    setup(**config)



#To prepare for model training, we import the necessary functions and submodules from keras
from keras.models import Sequential
from keras.layers.core import Dropout, Reshape, Dense, Activation, Flatten
from keras.layers.convolutional import Conv2D, MaxPooling2D
from keras.optimizers import Adadelta, SGD, RMSprop;
import keras.losses;
from keras.constraints import maxnorm;
from keras.layers.normalization import BatchNormalization
from keras.regularizers import l1, l2
from keras.callbacks import EarlyStopping, History
from keras import backend as K
K.set_image_data_format('channels_last')
from dragonn.callbacks import *

#Import matplotlib utilities for plotting grid search results: 
import matplotlib.pyplot as plt


def hyperparam_get_model(input_shape,
              num_tasks=1,
              num_layers=1,
              num_filters=10,
              kernel_size=15,
              pool_size=35):
    '''
    Performs hyperparameter search in accordance with Figure 6 in the DragoNN manuscript 

    input_shape: tuple with dimensions of input 
    num_tasks: number of prediction tasks, default=1 
    num_layers: number of 2D convolution layers, default=1 
    num_filters: number of filters to use in each 2D convolution layer, default=10 
    kernel_size: filter dimension for each 2D convolution layer, (1,n), default n=15  
    pool_size: 2D MaxPooling pool size, default=35 

    Returns a Sequential model with the specified parameters. 
    '''
    model=Sequential()
    for i in range(num_layers):
        model.add(Conv2D(filters=num_filters,
                         kernel_size=(1,kernel_size),
                         input_shape=input_shape))
        model.add(Activation("relu"))
    model.add(MaxPooling2D(pool_size=(1,pool_size)))
    model.add(Flatten())
    model.add(Dense(num_tasks))
    model.add(Activation("sigmoid"))
    model.compile(optimizer="adam",loss="binary_crossentropy")
    return model

def hyperparam_train_model(data,model,num_training_examples=None,epochs=150,patience=7):
    #train the model
    if num_training_examples==None:
        #use all training examples
        x=data.X_train
        y=data.y_train
    else:
        #use the specified number of training examples 
        x=data.X_train[0:num_training_examples]
        y=data.y_train[0:num_training_examples]
        
    history=model.fit(x=x,
                      y=y,
                      batch_size=128,
                      epochs=epochs,
                      verbose=0,
                      callbacks=[EarlyStopping(patience=patience),
                                 History()],
                      validation_data=(data.X_valid,data.y_valid))
    
    #get auPRC on test set
    test_predictions=model.predict(data.X_test)
    test_performance=ClassificationResult(data.y_test,test_predictions)
    return [i['auROC'] for i in test_performance.results]

def hyperparam_plot_test_auPRC(param_grid,auPRC_dict,xlabel="",ylabel=""):
    #define a list of colors for multi-tasked models 
    colors=["#000000","#1f78b4","#e31a1c","#33a02c","#ff7f00","#a6cee3","#cab2d6","#fdbf6f","#fb9a99","#b2df8a"]
    plt.figure(1,figsize=(20,5))
    datasets=list(auPRC_dict.keys())
    num_datasets=len(datasets)
    
    for i in range(num_datasets):
        cur_dataset=datasets[i]
        cur_auPRC=auPRC_dict[cur_dataset]
        
        #create subplot for current dataset
        plt.subplot(1,num_datasets,i+1)

        #make sure that the input is not an empty list (i.e. some performrance values recorded) 
        assert len(cur_auPRC)>0

        num_tasks=len(cur_auPRC[0])        
        #for purposes of the tutorial, we want to limit analysis to 10 tasks. 
        assert num_tasks < 11

        #add line for param value vs test auPRC for each task 
        for task_index in range(num_tasks):
            x=param_grid
            y=[entry[task_index] for entry in cur_auPRC]
            plt.plot(x,y,colors[task_index])
            plt.title(cur_dataset)
            plt.xlabel(xlabel)
            plt.ylabel(ylabel)
            plt.ylim(0.4,1)
    plt.subplots_adjust(hspace=0) 
    plt.show() 






from dragonn.utils import rolling_window
import numpy as np
from sklearn.metrics import auc, precision_recall_curve
import matplotlib.pyplot as plt
import pdb
from keras import backend as K
import tensorflow as tf

    

def positionalPRC(embeddings, scores,window_stride=1, coverage_thresh_for_positive=0.8,verbose=False):
    '''
    window_stride: number of bases to shift adjacent sequence windows by; default=1
    coverage_thresh_for_positive: sequence window must overlap the motif by this fraction (0 - 1) for the window to be labeled positive. 
    embeddings: the list of motif embeddings from simulation data (most likely from  simulation_data.valid_embeddings)
    scores: a 2-d numpy array (number of sequences x length of sequence)  Generally one of: motif scan scores, ISM scores, gradient x input scores, deepLIFT scores for each sequence in the dataset 

    returns: dictionary of motif_name-->[precision, recall, auPRC]
    '''
    #we concatenate across all sequences in the input dataset
    assert scores.shape[0]==len(embeddings)
    if len(scores.shape)>2:
        scores=np.squeeze(scores)
    
    #get the length of input sequences in the dataset 
    seq_length=scores.shape[1]

    #keep lists of labels and predictions for each embedded entity
    all_prc_inputs={} 

    #iterate through all input sequences 
    for i in range(len(embeddings)):
        seq_embeddings=embeddings[i]
        seq_scores=scores[i]
        seq_prc_inputs=dict() 
        
        #sequence may have multiple embeddings
        for embedding in seq_embeddings:
            motif_length=len(embedding.what.string)
            motif_name=embedding.what.stringDescription
            embedding_start_pos=embedding.startPos 
            
            if motif_name not in all_prc_inputs:
                all_prc_inputs[motif_name]=dict()
                all_prc_inputs[motif_name]['labels']=[]
                all_prc_inputs[motif_name]['scores']=[]

            if motif_name not in seq_prc_inputs:
                seq_prc_inputs[motif_name]=dict()
                seq_prc_inputs[motif_name]['scores']=np.sum(rolling_window(seq_scores,motif_length),axis=1)
                seq_prc_inputs[motif_name]['labels']=None
            #label the window that starts at the embedding start position with 1.
            tmp_label_array=np.zeros((seq_length,))
            tmp_label_array[embedding_start_pos:embedding_start_pos+motif_length]=1
            tmp_label_windows=np.sum(rolling_window(tmp_label_array,motif_length),axis=1)
            min_window_sum=coverage_thresh_for_positive*motif_length            
            #ambiguous windows are designated with 0.5 to allow for use of np.maximum below 
            tmp_label_windows[(tmp_label_windows>0) & (tmp_label_windows<min_window_sum)]=0.5
            tmp_label_windows[tmp_label_windows>=min_window_sum]=1 #positive
            
            if type(seq_prc_inputs[motif_name]['labels']) is not np.ndarray:
                seq_prc_inputs[motif_name]['labels']=tmp_label_windows
            else:
                seq_prc_inputs[motif_name]['labels']=np.maximum(seq_prc_inputs[motif_name]['labels'],tmp_label_windows)
        #update the dictionary of PRC inputs concatenated across sequences
        for motif_name in seq_prc_inputs.keys():
            #drop any ambiguous indices
            non_ambiguous_indices=np.where(seq_prc_inputs[motif_name]['labels']!=0.5)
            #print(max(seq_prc_inputs[motif_name]['labels']))
            if(verbose==True):
                print(motif_name)
                print("labels:"+str(seq_prc_inputs[motif_name]['labels']))
                print("scores:"+str(seq_prc_inputs[motif_name]['scores']))
            all_prc_inputs[motif_name]['labels']+=list(seq_prc_inputs[motif_name]['labels'][non_ambiguous_indices])
            all_prc_inputs[motif_name]['scores']+=list(seq_prc_inputs[motif_name]['scores'][non_ambiguous_indices])

    #calculate the PRC values and auPRC
    prc_values=dict()
    for motif_name in all_prc_inputs:
        labels=all_prc_inputs[motif_name]['labels']
        scores=all_prc_inputs[motif_name]['scores']
        #we normalize the scores to a 0-1 range for sklean prc function
        normalized_scores = (scores-min(scores))/(max(scores)-min(scores))
        precision, recall = precision_recall_curve(labels, normalized_scores)[:2]
        #sort
        sorted_prc=sorted(zip(recall,precision))
        recall=[prc_val[0] for prc_val in sorted_prc]
        precision=[prc_val[1] for prc_val in sorted_prc]
        auPRC=auc(recall,precision)
        prc_values[motif_name]=[recall,precision,auPRC]
    return prc_values
    

    



from __future__ import absolute_import, division, print_function
import matplotlib
import numpy as np
import os
import subprocess
import sys
import tempfile
matplotlib.use('pdf')
import matplotlib.pyplot as plt
from abc import abstractmethod, ABCMeta
from dragonn.metrics import ClassificationResult
from sklearn.svm import SVC as scikit_SVC
from sklearn.tree import DecisionTreeClassifier as scikit_DecisionTree
from sklearn.ensemble import RandomForestClassifier
from keras.models import load_model
from dragonn.runtime_metrics import *
from dragonn.custom_losses import * 
import warnings
warnings.filterwarnings('ignore')

def load_dragonn_model(model_string):
    custom_objects={"recall":recall,
                    "sensitivity":recall,
                    "specificity":specificity,
                    "fpr":fpr,
                    "fnr":fnr,
                    "fdr":fdr,
                    "precision":precision,
                    "f1":f1,
                    "spearman_corr":spearman_corr,
                    "ambig_binary_crossentropy":ambig_binary_crossentropy,
                    "ambig_mean_squared_error":ambig_mean_squared_error}    
    model=load_model(model_string,custom_objects=custom_objects)
    return model


class Model(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def __init__(self, **hyperparameters):
        pass

    @abstractmethod
    def train(self, X, y, validation_data):
        pass

    @abstractmethod
    def predict(self, X):
        pass

    def test(self, X, y):
        return ClassificationResult(y, self.predict(X))

    def score(self, X, y, metric):
        return self.test(X, y)[metric]


class SequenceDNN(Model):
    """
    Sequence DNN models.

    Parameters
    ----------
    seq_length : int, optional
        length of input sequence.
    keras_model : instance of keras.models.Sequential, optional
        seq_length or keras_model must be specified.
    num_tasks : int, optional
        number of tasks. Default: 1.
    num_filters : list[int] | tuple[int]
        number of convolutional filters in each layer. Default: (15,).
    conv_width : list[int] | tuple[int]
        width of each layer's convolutional filters. Default: (15,).
    pool_width : int
        width of max pooling after the last layer. Default: 35.
    L1 : float
        strength of L1 penalty.
    dropout : float
        dropout probability in every convolutional layer. Default: 0.
    verbose: int
        Verbosity level during training. Valida values: 0, 1, 2.

    Returns
    -------
    Compiled DNN model.
    """

    def __init__(self, seq_length=None, keras_model=None,
                 use_RNN=False, num_tasks=1,
                 num_filters=(15, 15, 15), conv_width=(15, 15, 15),
                 pool_width=35, GRU_size=35, TDD_size=15,
                 L1=0, dropout=0.0, num_epochs=100, verbose=1):
        from keras.models import Sequential
        from keras.layers.core import (
            Activation, Dense, Dropout, Flatten,
            Permute, Reshape
        )
        from keras.layers.convolutional import Convolution2D, MaxPooling2D
        from keras.layers.recurrent import GRU
        from keras.regularizers import l1
        self.num_tasks = num_tasks
        self.num_epochs = num_epochs
        self.verbose = verbose
        self.train_metrics = []
        self.valid_metrics = []
        if keras_model is not None and seq_length is None:
            self.model = keras_model
            self.num_tasks = keras_model.layers[-1].output_shape[-1]
        elif seq_length is not None and keras_model is None:
            self.model = Sequential()
            assert len(num_filters) == len(conv_width)
            for i, (nb_filter, nb_col) in enumerate(zip(num_filters, conv_width)):
                conv_height = 4 if i == 0 else 1
                self.model.add(Convolution2D(
                    nb_filter=nb_filter, nb_row=conv_height,
                    nb_col=nb_col, activation='linear',
                    init='he_normal', input_shape=(1, 4, seq_length),
                    W_regularizer=l1(L1), b_regularizer=l1(L1)))
                self.model.add(Activation('relu'))
                self.model.add(Dropout(dropout))
            self.model.add(MaxPooling2D(pool_size=(1, pool_width)))
            if use_RNN:
                num_max_pool_outputs = self.model.layers[-1].output_shape[-1]
                self.model.add(Reshape((num_filters[-1], num_max_pool_outputs)))
                self.model.add(Permute((2, 1)))
                self.model.add(GRU(GRU_size, return_sequences=True))
                self.model.add(TimeDistributedDense(TDD_size, activation='relu'))
            self.model.add(Flatten())
            self.model.add(Dense(output_dim=self.num_tasks))
            self.model.add(Activation('sigmoid'))
            self.model.compile(optimizer='adam', loss='binary_crossentropy')
        else:
            raise ValueError("Exactly one of seq_length or keras_model must be specified!")

    def train(self, X, y, validation_data, early_stopping_metric='Loss',
              early_stopping_patience=5, save_best_model_to_prefix=None):
        if y.dtype != bool:
            assert set(np.unique(y)) == {0, 1}
            y = y.astype(bool)
        multitask = y.shape[1] > 1
        if not multitask:
            num_positives = y.sum()
            num_sequences = len(y)
            num_negatives = num_sequences - num_positives
        if self.verbose >= 1:
            print('Training model (* indicates new best result)...')
        X_valid, y_valid = validation_data
        early_stopping_wait = 0
        best_metric = np.inf if early_stopping_metric == 'Loss' else -np.inf
        for epoch in range(1, self.num_epochs + 1):
            self.model.fit(X, y, batch_size=128, nb_epoch=1,
                           class_weight={True: num_sequences / num_positives,
                                         False: num_sequences / num_negatives}
                           if not multitask else None, verbose=self.verbose >= 2)
            epoch_train_metrics = self.test(X, y)
            epoch_valid_metrics = self.test(X_valid, y_valid)
            self.train_metrics.append(epoch_train_metrics)
            self.valid_metrics.append(epoch_valid_metrics)
            if self.verbose >= 1:
                print('Epoch {}:'.format(epoch))
                print('Train {}'.format(epoch_train_metrics))
                print('Valid {}'.format(epoch_valid_metrics), end='')
            current_metric = epoch_valid_metrics[early_stopping_metric].mean()
            if (early_stopping_metric == 'Loss') == (current_metric <= best_metric):
                if self.verbose >= 1:
                    print(' *')
                best_metric = current_metric
                best_epoch = epoch
                early_stopping_wait = 0
                if save_best_model_to_prefix is not None:
                    self.save(save_best_model_to_prefix)
            else:
                if self.verbose >= 1:
                    print()
                if early_stopping_wait >= early_stopping_patience:
                    break
                early_stopping_wait += 1
        if self.verbose >= 1:
            print('Finished training after {} epochs.'.format(epoch))
            if save_best_model_to_prefix is not None:
                print("The best model's architecture and weights (from epoch {0}) "
                      'were saved to {1}.arch.json and {1}.weights.h5'.format(
                    best_epoch, save_best_model_to_prefix))

    def predict(self, X):
        return self.model.predict(X, batch_size=128, verbose=False)

    def get_sequence_filters(self):
        """
        Returns 3D array of 2D sequence filters.
        """
        return self.model.layers[0].get_weights()[0].squeeze(axis=1)



    @staticmethod
    def _plot_scores(X, output_directory, peak_width, score_func, score_name):
        from dragonn.plot import plot_bases_on_ax
        scores = score_func(X).squeeze(axis=2)  # (num_task, num_samples, num_bases, sequence_length)
        try:
            os.makedirs(output_directory)
        except OSError:
            pass
        num_tasks = len(scores)
        for task_index, task_scores in enumerate(scores):
            for sequence_index, sequence_scores in enumerate(task_scores):
                # sequence_scores is num_bases x sequence_length
                basewise_max_sequence_scores = sequence_scores.max(axis=0)
                plt.clf()
                figure, (top_axis, bottom_axis) = plt.subplots(2)
                top_axis.plot(range(1, len(basewise_max_sequence_scores) + 1),
                              basewise_max_sequence_scores)
                top_axis.set_title('{} scores (motif highlighted)'.format(score_name))
                peak_position = basewise_max_sequence_scores.argmax()
                top_axis.axvspan(peak_position - peak_width, peak_position + peak_width,
                                 color='grey', alpha=0.1)
                peak_sequence_scores = sequence_scores[:, peak_position - peak_width :
                                                          peak_position + peak_width].T
                # Set non-max letter_heights to zero
                letter_heights = np.zeros_like(peak_sequence_scores)
                letter_heights[np.arange(len(letter_heights)),
                               peak_sequence_scores.argmax(axis=1)] = \
                    basewise_max_sequence_scores[peak_position - peak_width :
                                                 peak_position + peak_width]
                plot_bases_on_ax(letter_heights, bottom_axis)
                bottom_axis.set_xticklabels(tuple(map(
                    str, np.arange(peak_position - peak_width, peak_position + peak_width + 1))))
                bottom_axis.tick_params(axis='x', labelsize='small')
                plt.xlabel('Position')
                plt.ylabel('Score')
                plt.savefig(os.path.join(output_directory, 'sequence_{}{}'.format(
                    sequence_index, '_task_{}'.format(task_index) if num_tasks > 1 else '')))
                plt.close()

    def plot_deeplift(self, X, output_directory, peak_width=10):
        self._plot_scores(X, output_directory, peak_width,
                          score_func=self.deeplift, score_name='DeepLift')

    def plot_in_silico_mutagenesis(self, X, output_directory, peak_width=10):
        self._plot_scores(X, output_directory, peak_width,
                          score_func=self.in_silico_mutagenesis, score_name='ISM')

    def plot_architecture(self, output_file):
        from dragonn.visualize_util import plot as plot_keras_model
        plot_keras_model(self.model, output_file, show_shape=True)

    def save(self, save_best_model_to_prefix):
        arch_fname = save_best_model_to_prefix + '.arch.json'
        weights_fname = save_best_model_to_prefix + '.weights.h5'
        open(arch_fname, 'w').write(self.model.to_json())
        self.model.save_weights(weights_fname, overwrite=True)

    @staticmethod
    def load(model_hdf5_fname=None, arch_fname=None, weights_fname=None):
        if model_hdf5_fname!=None:
            from keras.models import load_model
            sequence_dnn=SequenceDNN(keras_model=load_model(model_hdf5_fname))
        else:
            from keras.models import model_from_json
            model_json_string = open(arch_fname).read()
            sequence_dnn = SequenceDNN(keras_model=model_from_json(model_json_string))
            if weights_fname is not None:
                sequence_dnn.model.load_weights(weights_fname)
        return sequence_dnn

class MotifScoreRNN(Model):

    def __init__(self, input_shape, gru_size=10, tdd_size=4):
        from keras.models import Sequential
        from keras.layers.core import (
            Activation, Dense, Flatten, TimeDistributedDense
        )
        from keras.layers.recurrent import GRU
        self.model = Sequential()
        self.model.add(GRU(gru_size, return_sequences=True,
                           input_shape=input_shape))
        if tdd_size is not None:
            self.model.add(TimeDistributedDense(tdd_size))
        self.model.add(Flatten())
        self.model.add(Dense(1))
        self.model.add(Activation('sigmoid'))
        print('Compiling model...')
        self.model.compile(optimizer='adam', loss='binary_crossentropy')

    def train(self, X, y, validation_data):
        from keras.callbacks import EarlyStopping
        print('Training model...')
        multitask = y.shape[1] > 1
        if not multitask:
            num_positives = y.sum()
            num_sequences = len(y)
            num_negatives = num_sequences - num_positives
        self.model.fit(
            X, y, batch_size=128, nb_epoch=100,
            validation_data=validation_data,
            class_weight={True: num_sequences / num_positives,
                          False: num_sequences / num_negatives}
            if not multitask else None,
            callbacks=[EarlyStopping(monitor='val_loss', patience=10)],
            verbose=True)

    def predict(self, X):
        return self.model.predict(X, batch_size=128, verbose=False)


class gkmSVM(Model):

    def __init__(self, prefix='./gkmSVM', word_length=11, mismatches=3, C=1,
                 threads=1, cache_memory=100, verbosity=4):
        self.word_length = word_length
        self.mismatches = mismatches
        self.C = C
        self.threads = threads
        self.prefix = '_'.join(map(str, (prefix, word_length, mismatches, C)))
        options_list = zip(
            ['-l', '-d', '-c', '-T', '-m', '-v'],
            map(str, (word_length, mismatches, C, threads, cache_memory, verbosity)))
        self.options = ' '.join([' '.join(option) for option in options_list])

    @property
    def model_file(self):
        model_fname = '{}.model.txt'.format(self.prefix)
        return model_fname if os.path.isfile(model_fname) else None

    @staticmethod
    def encode_sequence_into_fasta_file(sequence_iterator, ofname):
        """writes sequences into fasta file
        """
        with open(ofname, "w") as wf:
            for i, seq in enumerate(sequence_iterator):
                print('>{}'.format(i), file=wf)
                print(seq, file=wf)

    def train(self, X, y, validation_data=None):
        """
        Trains gkm-svm, saves model file.
        """
        y = y.squeeze()
        pos_sequence = X[y]
        neg_sequence = X[~y]
        pos_fname = "%s.pos_seq.fa" % self.prefix
        neg_fname = "%s.neg_seq.fa" % self.prefix
        # create temporary fasta files
        self.encode_sequence_into_fasta_file(pos_sequence, pos_fname)
        self.encode_sequence_into_fasta_file(neg_sequence, neg_fname)
        # run command
        command = ' '.join(
            ('gkmtrain', self.options, pos_fname, neg_fname, self.prefix))
        process = subprocess.Popen(command, stdout=subprocess.PIPE, shell=True)
        process.wait()  # wait for it to finish
        # remove fasta files
        os.system("rm %s" % pos_fname)
        os.system("rm %s" % neg_fname)

    def predict(self, X):
        if self.model_file is None:
            raise RuntimeError("GkmSvm hasn't been trained!")
        # write test fasta file
        test_fname = "%s.test.fa" % self.prefix
        self.encode_sequence_into_fasta_file(X, test_fname)
        # test gkmsvm
        temp_ofp = tempfile.NamedTemporaryFile()
        threads_option = '-T %s' % (str(self.threads))
        command = ' '.join(['gkmpredict',
                            test_fname,
                            self.model_file,
                            temp_ofp.name,
                            threads_option])
        process = subprocess.Popen(command, shell=True)
        process.wait()  # wait for it to finish
        os.system("rm %s" % test_fname)  # remove fasta file
        # get classification results
        temp_ofp.seek(0)
        y = np.array([line.split()[-1] for line in temp_ofp], dtype=float)
        temp_ofp.close()
        return np.expand_dims(y, 1)


class SVC(Model):

    def __init__(self):
        self.classifier = scikit_SVC(probability=True, kernel='linear')

    def train(self, X, y, validation_data=None):
        self.classifier.fit(X, y)

    def predict(self, X):
        return self.classifier.predict_proba(X)[:, 1:]


class DecisionTree(Model):

    def __init__(self):
        self.classifier = scikit_DecisionTree()

    def train(self, X, y, validation_data=None):
        self.classifier.fit(X, y)

    def predict(self, X):
        predictions = np.asarray(self.classifier.predict_proba(X))[..., 1]
        if len(predictions.shape) == 2:  # multitask
            predictions = predictions.T
        else:  # single-task
            predictions = np.expand_dims(predictions, 1)
        return predictions


class RandomForest(DecisionTree):

    def __init__(self):
        self.classifier = RandomForestClassifier(n_estimators=100)
